#! python
###########################################################################
# 
# File:            fms_pti.py
#
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Not allowed to be used for mass downloading
#                  of maps!
#
#
HELP_TEXT = """
# Module options:      --terrain   - Fetch terrain (maasto) maps instead of 
#                                    service default base maps
#
#                      --orto      - Fetch air pictures instead of 
#                                    service default base maps
#
#                      --hillshade - Fetch hillshade ("3D") pictures instead of 
#                                    service default base maps
"""
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         3.3a
#
# Date:            17.01.2022
#
# Functions:       -
#                   
# Description:     Map service class for Paikkatietoikkuna.
#
# Requirements:    Python interpreter 3.6 or newer (www.python.org)
#                  (tested with 3.10.1)
# 
#                  Python Imaging Library (tested with 1.1.5) 
#                  (www.pythonware.org/products/pil/)
#
# Version history: ** 01.10.2013 v3.2a (Olli Lammi) **
#                  First implementation.
#
#                  ** 22.02.2015 v3.2b (Olli Lammi) **
#                  Paikkatietoikkuna.fi service structure had changed.
#                  Rewritten the module to support new service structure.
#
#                  ** 05.12.2016 v3.2e (Olli Lammi) **
#                  Added new layers. Added --hillshade material.
#
#                  ** 17.01.2022 v3.3a (Olli Lammi) **
#                  Python3 version.
#
###########################################################################

# Imports

import sys, os
import math, re, os.path
import tempfile, shutil
import random

import urllib.request

from PIL import Image  # Python Imaging Library 

import coordinates  # KKJ to and from WGS84 coordinate transforms

import fms

###########################################################################

# Constants

MOD_VERSION = 'v3.3a'
SERVICE_CODE = 'pti'
CACHE_CODE = 'pti'

# constants for the picture and distance sizes and ratios
PIC_ORIG_SIZEX = 256
PIC_ORIG_SIZEY = 256
PIC_SIZEX = PIC_ORIG_SIZEX
PIC_SIZEY = PIC_ORIG_SIZEY


PIC_SCALES = { '1:1000': [15, 0.25, 'JHS180 level 15'], \
               '1:2000': [14, 0.5, 'JHS180 level 14'], \
               '1:4000': [13, 1.0, 'JHS180 level 13'], \
               '1:8000': [12, 2.0, 'JHS180 level 12'], \
               '1:16000': [11, 4.0, 'JHS180 level 11'], \
               '1:32000': [10, 8.0, 'JHS180 level 10'], \
               '1:64000': [9, 16.0, 'JHS180 level 9'],\
               '1:128000': [8, 32.0, 'JHS180 level 8'],\
               '1:256000': [7, 64.0, 'JHS180 level 7'],\
               '1:512000': [6, 128.0, 'JHS180 level 6'],\
               '1:1024000': [5, 256.0, 'JHS180 level 5'],\
               '1:2048000': [4, 512.0, 'JHS180 level 4'],\
               '1:4096000': [3, 1024.0, 'JHS180 level 3'] }

PIC_SERVICES = { \
               'DEFAULT': {'path': 'maasto', 'layer': 'taustakartta', 'id': 'default', 'ext': 'png'}, \
               'terrain': {'path': 'maasto', 'layer': 'maastokartta', 'id': 'default', 'ext': 'png'}, \
               'hillshade': {'path': 'maasto', 'layer': 'korkeusmalli_vinovalo', 'id': 'default', 'ext': 'png'}, \
               'orto': {'path': 'maasto', 'layer': 'ortokuva', 'id': 'default', 'ext': 'jpg'} \
               }

PIC_SCALE_DEFAULT = '1:16000'

LAYERS = { 'kiinteistotunnukset': ['Kiinteistötunnukset', 1.0], \
           'kiinteistojaotus': ['Kiinteistöjaotus', 1.0], \
           'rakennukset': ['Rakennukset', 1.0], \
           'maankaytto': ['Maankaytto', 0.7], \
           'maanpeite': ['Maanpeite', 0.7], \
           'hydrografia': ['Hydrografia', 1.0], \
           'korkeus': ['Korkeuskäyrät', 1.0], \
           'paikannimet': ['Paikannimet', 1.0], \
           'osoitteet': ['Osoitteet', 1.0], \
           'liikenneverkot': ['Liikenneverkot', 1.0], \
           'kuntarajat': ['Kuntarajat', 1.0], \
           'kuntatunnukset': ['Kuntatunnukset', 1.0], \
           'hallyksikot': ['Hallinnolliset yksiköt', 1.0], \
           'tm35lehtijako': ['TM35-lehtijako', 1.0] \
         }

LAYER_SERVICES = { \
               'kiinteistotunnukset': {'path': 'kiinteisto', 'layer': 'kiinteistotunnukset', 'id': 'default', 'ext': 'png'}, \
               'kiinteistojaotus': {'path': 'kiinteisto', 'layer': 'kiinteistojaotus', 'id': 'default', 'ext': 'png'}, \
               'rakennukset': {'path': 'teema', 'layer': 'rakennukset', 'id': 'default', 'ext': 'png'}, \
               'maankaytto': {'path': 'teema', 'layer': 'maankaytto', 'id': 'default', 'ext': 'png'}, \
               'maanpeite': {'path': 'teema', 'layer': 'maanpeite', 'id': 'default', 'ext': 'png'}, \
               'hydrografia': {'path': 'teema', 'layer': 'hydrografia', 'id': 'default', 'ext': 'png'}, \
               'korkeus': {'path': 'teema', 'layer': 'korkeus', 'id': 'default', 'ext': 'png'}, \
               'paikannimet': {'path': 'teema', 'layer': 'paikannimet', 'id': 'default', 'ext': 'png'}, \
               'osoitteet': {'path': 'teema', 'layer': 'osoitteet', 'id': 'default', 'ext': 'png'}, \
               'liikenneverkot': {'path': 'teema', 'layer': 'liikenneverkot', 'id': 'default', 'ext': 'png'}, \
               'kuntarajat': {'path': 'teema', 'layer': 'kuntarajat', 'id': 'default', 'ext': 'png'}, \
               'kuntatunnukset': {'path': 'teema', 'layer': 'kuntatunnukset', 'id': 'default', 'ext': 'png'}, \
               'hallyksikot': {'path': 'teema', 'layer': 'hallinnolliset_yksikot', 'id': 'default', 'ext': 'png'}, \
               'tm35lehtijako': {'path': 'teema', 'layer': 'TM35_lehtijako', 'id': 'default', 'ext': 'png'} \
                 }

LAYERORDER = ['kiinteistotunnukset', 'kiinteistojaotus', 'rakennukset', 'maankaytto', 'maanpeite', 'hydrografia', 'korkeus', \
              'paikannimet', 'osoitteet', 'liikenneverkot', 'kuntarajat', 'kuntatunnukset', 'hallyksikot', 'tm35lehtijako']


###########################################################################

# Classes

class FMSPaikkatietoikkuna(fms.FetchMapService):
    def __init__(self):
        self.SERVICE_CODE = SERVICE_CODE
        self.CACHE_CODE = CACHE_CODE  
          
        self.TERRAIN = 0  
        self.HILLSHADE = 0  
        self.ORTO = 0  
          
        self.PIC_SCALE_DEFAULT = PIC_SCALE_DEFAULT
        self.SetScale(self.PIC_SCALE_DEFAULT)
     
    
    ###########################################################################
    # Function:    GetVersion
    ###########################################################################
    # Description: Returns the module version string. 
    ###########################################################################    
    def GetVersion(self):
        return MOD_VERSION
    
    
    ###########################################################################
    # Function:    HandleParameter
    ###########################################################################
    # Description: Handles command line parameter. Returns 0, if the parameter
    #              was invalid. Returns 1 if the parameter was ok. 
    ###########################################################################  
    def HandleParameter(self, param):    
        if param == '--terrain':
            self.TERRAIN = 1
        elif param == '--orto':
            self.ORTO = 1
        elif param == '--hillshade':
            self.HILLSHADE = 1
        else:
            return 0
        
        return 1


    ###########################################################################
    # Function:    SetScale
    ###########################################################################
    # Description: Initializes the service and values to retrieve map parts
    #              in given scale. The parameter must be one of the values
    #              returned by GetScales.
    ###########################################################################
    def SetScale(self, scale):
        # determine picture scale and ratio
        try:
            self.PIC_SIZEX = PIC_SIZEX
            self.PIC_SIZEY = PIC_SIZEY
            self.SCALE = scale
            self.SCALEVAL = PIC_SCALES[scale][0]
            self.PIC_RATIO = PIC_SCALES[scale][1]
            self.MAP_SIZEX = self.PIC_SIZEX * self.PIC_RATIO
            self.MAP_SIZEY = self.PIC_SIZEY * self.PIC_RATIO
        except:
            return 0
        return 1


    ###########################################################################
    # Function:    GetScales
    ###########################################################################
    # Description: Returns list of scales that the service supports
    ###########################################################################    
    def GetScales(self):
        scales = []
        for sc in PIC_SCALES.keys():
            scales.append([PIC_SCALES[sc], sc])
        scales.sort()
        scales.reverse()
        scales2 = []
        for sc in scales:
            scales2.append([sc[1], sc[0][2], sc[0][1]])
        return scales2
    
    
    ###########################################################################
    # Function:    GetPicIndexes
    ###########################################################################
    # Description: Returns two (x,y) tuples with indexes for map image that has
    #              the given (mapx, mapy) point and the lower left xy
    #              coordinates of the indexed map picture.
    ###########################################################################
    def GetPicIndexes(self, mapx, mapy):
        xcenter = math.pow(2.0, self.SCALEVAL - 1)
        tilex = int(math.floor(xcenter + (mapx + (0.25 * self.PIC_RATIO) - 500000.0) / self.MAP_SIZEX))
        tiley = int(math.floor((8388608.0 - (mapy + (0.25 * self.PIC_RATIO))) / self.MAP_SIZEY))

        (fetchx, fetchy) = self.GetTileLowerLeft(tilex, tiley)
                                    
        return ((tilex, tiley), (fetchx, fetchy))


    def GetTileLowerLeft(self, tilex, tiley):
        xcenter = math.pow(2.0, self.SCALEVAL - 1)
        llx = int(math.floor(((tilex - xcenter) * self.MAP_SIZEX) + 500000.0))
        lly = int(math.floor(8388608.0 - ((tiley + 1) * self.MAP_SIZEY)))
        return (llx, lly)

    
    ###########################################################################
    # Method:      FetchPic
    ###########################################################################
    # Description: Fetches one map tile from Paikkatietoikkuna.fi.
    #              Result file will saved in cachedir. Tempdir is used as temp directory  
    #              while downloading.
    ###########################################################################

    def FetchPic(self, fetchr, fetchc, cachedir, tempdir, layer):
        fname = cachedir + self.GetCacheName(fetchc, fetchr, layer)
                
        tempfname = None
        waitsize_x = PIC_ORIG_SIZEX
        waitsize_y = PIC_ORIG_SIZEY

        url = 'https://karttamoottori.maanmittauslaitos.fi/'

        if len(layer) > 0:
            serviceset = LAYER_SERVICES[layer]
        else:        
            if self.TERRAIN:
                serviceset = PIC_SERVICES['terrain']
            elif self.ORTO:
                serviceset = PIC_SERVICES['orto']
            elif self.HILLSHADE:
                serviceset = PIC_SERVICES['hillshade']
            else:            
                serviceset = PIC_SERVICES['DEFAULT']
        
        url = url + "%s/wmts/1.0.0/%s/%s/ETRS-TM35FIN/" % (serviceset['path'], serviceset['layer'], serviceset['id'])    
        url = url + "%d/%d/%d.%s" % (self.SCALEVAL, fetchr, fetchc, serviceset['ext'])
 
        USERAGENT = "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.0.4) Gecko/20060508 Firefox/1.5.0.4"
  
        res = ''
        try:
          req = urllib.request.Request(url)
          req.add_header("User-Agent", USERAGENT)
  
          f = urllib.request.urlopen(req)
          res = f.read()
          f.close()
        except:
          return 0
   
        tempfname = ''
        # create a temporary file name for the map
        try:
            (tempfd, tempfname) = tempfile.mkstemp('.png', 'tmpmap_', tempdir)
            os.close(tempfd)
        except:
            print('ERROR: Cannot create temporary file for map image file')
            return 0

        try:
            f = open(tempfname,'wb')
            f.write(res)
            f.close()
        except:
            print("ERROR: Cannot write temporary map image file: " + tempfname)
            return 0
    
        try:
            im = Image.open(tempfname)
            im.load()
    
            if im.size[0] != waitsize_x or im.size[1] != waitsize_y:
                print("ERROR: Retrieved image is of invalid size (received: %dx%d, expected: %dx%d)" % (im.size[0], im.size[1], waitsize_x, waitsize_y))
                try:
                    os.unlink(tempfname)
                except:
                    pass
                return 0 
                    
            im.save(tempfname)
        except Exception as e:
            print("ERROR: Cannot convert map image or rewrite temporary map image file: " + tempfname)
            print(e.__str__())
            try:
                os.unlink(tempfname)
            except:
                pass
            return 0

        if tempfname == None:
            return 0
 
        # check if the map cache hash directory is created
        dname = os.path.dirname(fname)
        if not os.path.isdir(dname):
            try:
                os.makedirs(dname)
            except:
                print('ERROR: Cannot create map image directory: ' + dname)
                try:
                    os.unlink(tempfname)
                except:
                    pass
                return 0
    
        # move the temporary map file to the cache
        try:
            shutil.move(tempfname, fname)  
        except:
            print('ERROR: Cannot move loaded temporary map image to map cache.')
            try:
                os.unlink(tempfname)
            except:
                pass
            return 0
  
        return 1
         
         
    ###########################################################################
    # Method:    GetCacheName
    ###########################################################################
    # Description: Generates a name for the map image in cache according to
    #              location and other parameters.
    ###########################################################################
    def GetCacheName(self, x, y, layer):
        fname = "%s_%d_%d" % (self.SCALEVAL, int(self.PIC_SIZEX), int(self.PIC_SIZEY))
  
        if len(layer) > 0:
            fname = fname + '_' + layer
        elif self.TERRAIN:
            fname = fname + '_terrain'
        elif self.ORTO:
            fname = fname + '_orto'
        elif self.HILLSHADE:
            fname = fname + '_hillshade'

        (hashx, hashy) = self.GetTileLowerLeft(x, y)
        hash = "%d_%d" % (HashRound(hashy), HashRound(hashx))
        fname = fname + '/' + hash
  
        fname = fname + "/map_%s_%s" % (self.CACHE_CODE, self.SCALEVAL)  
        
        fname = fname + '_'+ repr(int(round(y))) + '_' + repr(int(round(x))) + '.png'

        return fname


    ###########################################################################
    # Method:      PostProcess
    ###########################################################################
    # Description: Post processes the entire constructed map image according
    #              to service requirements.
    ###########################################################################    
    def PostProcess(self, im):
        pass


    ###########################################################################
    # Method:      TranslateToServiceCoordinatesXY
    ###########################################################################
    # Description: Does service require coordinates to be transferred to YKJ?
    #              Returns: tuple (type, x, y) where
    #              type - type string of the coordinate system
    #              x - easting or lon coordinate 
    #              y - northing or lat coordinate
    ###########################################################################       
    def TranslateToServiceCoordinatesXY(self, type, x, y):
        if (type != coordinates.COORD_TYPE_ETRSTM35FIN):
            temp = coordinates.Translate({'type': type, 'N': y, 'E': x}, \
                                         coordinates.COORD_TYPE_ETRSTM35FIN)
            return (temp['type'], temp['E'], temp['N'])
        return (type, x, y)

    
    ###########################################################################
    # Method:      GetHelp
    ###########################################################################
    # Description: Get service module specific help string.
    ###########################################################################    
    def GetHelp(self):
        return HELP_TEXT


    ###########################################################################
    # Method:      GetServiceDelayLimits
    ###########################################################################
    # Description: Get service specific delay times as tuple (min, max) in sec.
    ###########################################################################               
    def GetServiceDelayLimits(self):
        return (2, 6)
        

    ###########################################################################
    # Method:      GetAvailableLayers
    ###########################################################################
    # Description: Get list of available layer names. 
    ###########################################################################               
    def GetAvailableLayers(self):
        return LAYERORDER


    ###########################################################################
    # Method:      GetLayerInfo
    ###########################################################################
    # Description: Get info for given layer (list of layer name and opacity). 
    ###########################################################################               
    def GetLayerInfo(self, layer):
        return [LAYERS[layer][0], LAYERS[layer][1]]
    
            
    ###########################################################################
    # Method:      IsSelfCaching
    ###########################################################################
    # Description: Should return true, if module is self caching and
    #              GetCachePic will provide the image from the cache. 
    ###########################################################################   
    def IsSelfCaching(self):
        return 0
        
###########################################################################

# Functions

def HashRound(x):
    return ((x / 10000) / 5) * 5 


